package bank1.basic;

import java.util.List;

/**
 * System simulation. Access point of all related information.
 */
public class Simulation {
	private Bank bank;
	private StatisticManager statisticManager;
	private SimulationEntry systemEntry;

	public Simulation(SimulationEntry simulationEntry) {
		this.systemEntry = simulationEntry;
		int cashierCount = simulationEntry.getCashierCount();
		bank = new Bank(cashierCount);
		statisticManager = new StatisticManager();
	}

	/**
	 * This is the main simulation process, that products simulation results. 
	 */
	public void simulate() {
		int simulationDuration = systemEntry.getSimulationDuration();
		int clientArrivalInterval = systemEntry.getClientArrivalInterval();
		
		for (int currentSystemTime = 0; currentSystemTime < simulationDuration; currentSystemTime++) {

			statisticManager.simulationDurationRecord();
			SimulationUtility.printBankStat(currentSystemTime, bank);

			updateBank(currentSystemTime);

			boolean newClientArrival = newClientArrival(clientArrivalInterval, currentSystemTime);
			if (newClientArrival) {

				Client client = new Client(currentSystemTime);

				Cashier freeCashier = bank.getFreeCashier();
				if (freeCashier == null) {
					SimulationUtility.printClientArrival(false);
					Queue queue = bank.getQueue();
					queue.addQueueLast(client);
				} else {
					SimulationUtility.printClientArrival(true);
					serveClient(currentSystemTime, freeCashier, client);
				}
			}
		}
	}

	private void updateBank(int currentSystemTime) {
		List<Cashier> cashiers = bank.getCashiers();
		for (Cashier cashier : cashiers) {

			if (!cashier.isFree()) {
				statisticManager.cashierOccupationRecord();
			}

			cashier.work();

			if (cashier.serviceFinished()) {
				Client leavingClient = cashier.getServingClient();
				leavingClient.setDepartureTime(currentSystemTime);
				SimulationUtility.printClientDeparture();
				statisticManager.register(leavingClient);

				cashier.setServingClient(null);

				Queue queue = bank.getQueue();
				if (!queue.isEmpty()) {
					Client waitingClient = queue.getQueueFirst();
					serveClient(currentSystemTime, cashier, waitingClient);
				}
			}
		}
	}

	private void serveClient(int currentSystemTime, Cashier cashier, Client client) {
		int randomServiceTime = generateRandomServiceTime();
		client.setServiceStartTime(currentSystemTime);
		cashier.serve(client, randomServiceTime);
		SimulationUtility.printServiceTimeTrace(currentSystemTime, randomServiceTime);
	}

	private int generateRandomServiceTime() {
		int minServiceTime = systemEntry.getMinServiceTime();
		int maxServiceTime = systemEntry.getMaxServiceTime();
		int randomServiceTime = SimulationUtility.getRandomServiceTime(minServiceTime, maxServiceTime);
		return randomServiceTime;
	}

	private boolean newClientArrival(int clientArrivalInterval, int simIter) {
		return simIter % clientArrivalInterval == 0;
	}

	public String simulationResults() {
		StringBuffer results = new StringBuffer();
		results.append("########## Simulation results : ##########\n");
		results.append("Simulation Duration : " + statisticManager.getSimulationDuration() + "\n");
		results.append("Served client count : " + statisticManager.servedClientCount() + "\n");
		results.append("Average client waiting time : " + statisticManager.calculateAverageClientWaitingTime() + " \n");
		results.append("Average client service time : " + statisticManager.calculateAverageClientServiceTime() + " \n");
		results.append("Cashier occupation rate : " + statisticManager.calculateAverageCashierOccupationRate(systemEntry.getCashierCount()) + " %");
		return results.toString();
	}
}
